import dataclasses
import enum
import typing


class AlphabetType(enum.Enum):
    # bio
    amino = 'amino'
    DNA = 'DNA'
    RNA = 'RNA'
    # other
    coins = 'coins'
    dice = 'dice'
    custom = 'custom'


@dataclasses.dataclass
class Metadata:
    version_identifier: str
    model_name: str
    length: int
    alphabet_type: AlphabetType
    alphabet: typing.List[str]
    consensus_residue_annotation: bool

    accession_number: typing.Optional[str] = None
    description: typing.Optional[str] = None
    max_instance_length: typing.Optional[int] = None
    reference_annotation: typing.Optional[str] = None
    model_masked: bool = False
    consensus_structure_annotation: bool = False
    map_annotation: bool = False
    date: typing.Optional[str] = None
    command_line_log: typing.Optional[typing.List[str]] = None
    sequence_number: typing.Optional[int] = None
    effective_sequence_number: typing.Optional[float] = None
    checksum: typing.Optional[int] = None
    gathering_threshold: typing.Optional[typing.Tuple[float, float]] = None
    trusted_cutoff: typing.Optional[typing.Tuple[float, float]] = None
    noise_cutoff: typing.Optional[typing.Tuple[float, float]] = None
    statistical_parameters: typing.Optional[typing.List[typing.Tuple[str, str, float, float]]] = None
    build_command: typing.Optional[str] = None
    search_command: typing.Optional[str] = None


@dataclasses.dataclass
class BaseStep:
    p_emission_to_emission: float
    p_emission_to_insertion: float
    p_emission_to_deletion: float
    p_insertion_to_emission: float
    p_insertion_to_insertion: float
    p_deletion_to_emission: float
    p_deletion_to_deletion: float
    p_emission_char: typing.List[float]
    p_insertion_char: typing.List[float]


@dataclasses.dataclass
class Step(BaseStep):
    alignment_column_index: typing.Optional[int] = None
    consensus_residue_annotation: typing.Optional[str] = None
    reference_annotation: typing.Optional[str] = None
    mask_value: typing.Optional[str] = None
    annotation: typing.Optional[str] = None


@dataclasses.dataclass
class StartStep(BaseStep):
    pass


class HMM:
    def __init__(self, metadata: Metadata, steps: typing.List[Step], start_step: typing.Optional[StartStep] = None):
        self.metadata = metadata
        self.start_step = start_step
        self.steps = steps

    def check_equal_states(self, minimum_equality: float, how_many_min_chars: int = 2) -> int:
        result = 0
        for step in self.steps:
            if sum([p >= minimum_equality for p in step.p_emission_char]) >= how_many_min_chars:
                result += 1
        return result



import math
import pathlib
import typing

from hmm_profile import models


def save_many_to_file(
        hmms: typing.List[models.HMM],
        output: typing.Union[pathlib.Path, str],
) -> typing.Union[pathlib.Path, str]:
    with open(output, 'w') as f:
        for hmm in hmms:
            f.writelines(get_lines(hmm))
    return output


def save_to_file(hmm: models.HMM, output: typing.Union[pathlib.Path, str]) -> typing.Union[pathlib.Path, str]:
    with open(output, 'w') as f:
        f.writelines(get_lines(hmm))
    return output


def get_lines(hmm: models.HMM) -> typing.Generator[str, None, None]:
    yield from metadata_lines(hmm.metadata)
    yield from model_headers()
    if hmm.start_step is not None:
        yield from compo_lines(hmm.start_step)
    for i, step in enumerate(hmm.steps, start=1):
        yield from step_lines(i, step)
    yield from last_line()


def metadata_lines(metadata: models.Metadata) -> typing.Generator[str, None, None]:  # noqa: C901
    yield f'{metadata.version_identifier}\n'
    yield f'NAME  {metadata.model_name}\n'
    if metadata.accession_number is not None:
        yield f'ACC   {metadata.accession_number}\n'
    if metadata.description is not None:
        yield f'DESC  {metadata.description}\n'
    yield f'LENG  {metadata.length}\n'
    yield f'ALPH  {metadata.alphabet_type.value}\n'
    if metadata.reference_annotation is not None:
        yield f'RF    {metadata.reference_annotation}\n'
    if metadata.model_masked is not None:
        yield f'MM    {bool_to_str(metadata.model_masked)}\n'
    yield f'CONS  {bool_to_str(metadata.consensus_residue_annotation)}\n'
    if metadata.consensus_structure_annotation is not None:
        yield f'CS    {bool_to_str(metadata.consensus_structure_annotation)}\n'
    if metadata.map_annotation is not None:
        yield f'MAP   {bool_to_str(metadata.map_annotation)}\n'
    if metadata.date is not None:
        yield f'DATE  {metadata.date}\n'
    if metadata.sequence_number is not None:
        yield f'NSEQ  {metadata.sequence_number}\n'
    if metadata.effective_sequence_number is not None:
        yield f'EFFN  {metadata.effective_sequence_number}\n'
    if metadata.checksum is not None:
        yield f'CKSUM {metadata.checksum}\n'
    if metadata.gathering_threshold:
        a, b = metadata.gathering_threshold
        yield f'GA    {a:.2f} {b:.2f};\n'
    if metadata.trusted_cutoff:
        a, b = metadata.trusted_cutoff
        yield f'TC    {a:.2f} {b:.2f};\n'
    if metadata.noise_cutoff:
        a, b = metadata.noise_cutoff
        yield f'NC    {a:.2f} {b:.2f};\n'
    # TODO:
    # yield f'BM    \n'
    # yield f'SM    \n'
    # yield f'STATS \n'
    # yield f'STATS \n'
    # yield f'STATS \n'


def model_headers() -> typing.Generator[str, None, None]:
    yield ('HMM          A        C        D        E        F        G        H        I        K        '
           'L        M        N        P        Q        R        S        T        V        W        Y\n')
    yield '            m->m     m->i     m->d     i->m     i->i     d->m     d->d\n'


def compo_lines(start_step: models.StartStep) -> typing.Generator[str, None, None]:
    state_switch_values = ''.join([
        float_to_str(convert_probability(start_step.p_emission_to_emission)),
        float_to_str(convert_probability(start_step.p_emission_to_insertion)),
        float_to_str(convert_probability(start_step.p_emission_to_deletion)),
        float_to_str(convert_probability(start_step.p_insertion_to_emission)),
        float_to_str(convert_probability(start_step.p_insertion_to_insertion)),
        float_to_str(convert_probability(start_step.p_deletion_to_emission)),
        float_to_str(convert_probability(start_step.p_deletion_to_deletion)),
    ])
    insertion_values = ''.join([float_to_str(convert_probability(x)) for x in start_step.p_insertion_char])
    emission_values = ''.join([float_to_str(convert_probability(x)) for x in start_step.p_emission_char])
    yield f'  COMPO {emission_values}\n'
    yield f'        {insertion_values}\n'
    yield f'        {state_switch_values}\n'


def step_lines(line_number: int, step: models.Step) -> typing.Generator[str, None, None]:
    state_switch_values = ''.join([
        float_to_str(convert_probability(step.p_emission_to_emission)),
        float_to_str(convert_probability(step.p_emission_to_insertion)),
        float_to_str(convert_probability(step.p_emission_to_deletion)),
        float_to_str(convert_probability(step.p_insertion_to_emission)),
        float_to_str(convert_probability(step.p_insertion_to_insertion)),
        float_to_str(convert_probability(step.p_deletion_to_emission)),
        float_to_str(convert_probability(step.p_deletion_to_deletion)),
    ])
    emission_values = ''.join([float_to_str(convert_probability(x)) for x in step.p_emission_char])
    insertion_values = ''.join([float_to_str(convert_probability(x)) for x in step.p_insertion_char])
    yield f'{line_number:>7} {emission_values}\n'
    yield '        {}\n'.format(insertion_values)
    yield '        {}\n'.format(state_switch_values)


def float_to_str(number: typing.Union[float, str]) -> str:
    if isinstance(number, str):
        return f'{number:>9}'
    else:
        # dirty hack, to prevent negative zero...
        number += 0
        return f'{number:>9.5f}'


def convert_probability(p: float) -> typing.Union[str, float]:
    if p == 0:
        return '*'
    else:
        return -math.log(p)


def last_line() -> typing.Generator[str, None, None]:
    yield '//\n'


def bool_to_str(value: bool) -> str:
    return 'yes' if value else 'no'



#!/usr/bin/env python
import sys
from os import path

from setuptools import find_packages
from setuptools import setup

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

test_requirements = []
with open(path.join(this_directory, 'test_requirements.txt')) as f:
    for line in f:
        require = line.split('#', 1)[0].strip()
        if require:
            test_requirements.append(require)

if sys.version_info < (3, 7):
    install_requires = ['dataclasses']
else:
    install_requires = []

setup(
    name='hmm_profile',
    packages=find_packages(exclude=['tests']),
    version='0.0.5',
    url='https://github.com/Behoston/hmm_profile',
    license='MIT',
    author='Behoston',
    author_email='mlegiecki@gmail.com',
    description='Hidden Markov Model profile tools (reader/writer/data structures)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    py_modules=['hmm_profile'],
    include_package_data=True,
    install_requires=install_requires,
    tests_require=test_requirements,
    setup_requires=['pytest-runner'] if needs_pytest else [],
    platforms='any',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Other Environment',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Education',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development',
        'Typing :: Typed',
    ],
)



import logging
import math
import typing

from hmm_profile import models

logger = logging.getLogger(__name__)


def read_all(f: typing.TextIO) -> typing.Generator[models.HMM, None, None]:
    try:
        while True:
            yield read_single(f)
    except EOFError:
        return


def read_single(f: typing.TextIO) -> models.HMM:
    metadata = parse_metadata(f)
    steps = list(parse_steps(f, metadata.alphabet))
    start_step = None
    if isinstance(steps[0], models.StartStep):
        start_step = steps[0]
        steps = steps[1:]
    return models.HMM(metadata, steps, start_step)


def parse_metadata(f) -> models.Metadata:  # noqa: C901
    version = _parse_version(f)
    metadata_dict = {'version_identifier': version}
    statistical_parameters = []
    for line in f:
        tag, unparsed_value = line.strip().split(maxsplit=1)
        if tag == 'HMM':
            metadata_dict['alphabet'] = unparsed_value.split()
            f.readline()
            break
        elif tag == 'NAME':
            metadata_dict['model_name'] = unparsed_value
        elif tag == 'ACC':
            metadata_dict['accession_number'] = unparsed_value
        elif tag == 'DESC':
            metadata_dict['description'] = unparsed_value
        elif tag == 'LENG':
            metadata_dict['length'] = int(unparsed_value)
        elif tag == 'MAXL':
            metadata_dict['max_instance_length'] = int(unparsed_value)
        elif tag == 'ALPH':
            metadata_dict['alphabet_type'] = models.AlphabetType(unparsed_value)
        elif tag == 'RF':
            metadata_dict['reference_annotation'] = _parse_bool(unparsed_value)
        elif tag == 'MM':
            metadata_dict['model_masked'] = _parse_bool(unparsed_value)
        elif tag == 'CONS':
            metadata_dict['consensus_residue_annotation'] = _parse_bool(unparsed_value)
        elif tag == 'CS':
            metadata_dict['consensus_structure_annotation'] = _parse_bool(unparsed_value)
        elif tag == 'MAP':
            metadata_dict['map_annotation'] = _parse_bool(unparsed_value)
        elif tag == 'DATE':
            metadata_dict['date'] = unparsed_value
        elif tag == 'COM':
            if 'command_line_log' not in metadata_dict:
                metadata_dict['command_line_log'] = [unparsed_value]
            else:
                metadata_dict['command_line_log'].append(unparsed_value)
        elif tag == 'NSEQ':
            metadata_dict['sequence_number'] = int(unparsed_value)
        elif tag == 'EFFN':
            metadata_dict['effective_sequence_number'] = float(unparsed_value)
        elif tag == 'CKSUM':
            metadata_dict['checksum'] = int(unparsed_value)
        elif tag == 'GA':
            n1, n2 = unparsed_value.strip(';').split()
            metadata_dict['gathering_threshold'] = (float(n1), float(n2))
        elif tag == 'TC':
            n1, n2 = unparsed_value.strip(';').split()
            metadata_dict['trusted_cutoff'] = (float(n1), float(n2))
        elif tag == 'NC':
            n1, n2 = unparsed_value.strip(';').split()
            metadata_dict['noise_cutoff'] = (float(n1), float(n2))
        elif tag == 'STATS':
            s1, s2, f1, f2 = unparsed_value.strip(';').split()
            statistical_parameters.append((s1, s2, float(f1), float(f2)))
        elif tag == 'BM':
            metadata_dict['build_command'] = unparsed_value
        elif tag == 'SM':
            metadata_dict['search_command'] = unparsed_value
        else:
            logger.warning(f"Unknown tag '{tag}'!")
        if statistical_parameters:
            metadata_dict['statistical_parameters'] = statistical_parameters
    return models.Metadata(**metadata_dict)


def _parse_version(f: typing.TextIO) -> str:
    version = '\n'
    while version == '\n':
        version = f.readline()
    if version == '':
        raise EOFError
    else:
        return version.strip()


def _parse_bool(value: str) -> bool:
    value = value.lower()
    if value == 'yes':
        return True
    elif value == 'no':
        return False
    else:
        raise Exception(f"Cannot parse bool value: {value}.")


def parse_steps(f, alphabet: typing.List[str]) -> typing.Generator[models.Step, None, None]:
    alphabet_length = len(alphabet)
    while True:
        emission_line = f.readline().strip()
        if emission_line.startswith('//'):
            return

        emission_line = emission_line.split()
        insertion_line = f.readline().strip().split()
        state_switch_line = f.readline().strip().split()

        if emission_line[0] == 'COMPO':
            assert len(emission_line) == alphabet_length + 1, emission_line
            assert len(insertion_line) == alphabet_length, insertion_line
            assert len(state_switch_line) == 7, state_switch_line
            yield models.StartStep(
                p_emission_to_emission=convert_probability(state_switch_line[0]),
                p_emission_to_insertion=convert_probability(state_switch_line[1]),
                p_emission_to_deletion=convert_probability(state_switch_line[2]),
                p_insertion_to_emission=convert_probability(state_switch_line[3]),
                p_insertion_to_insertion=convert_probability(state_switch_line[4]),
                p_deletion_to_emission=convert_probability(state_switch_line[5]),
                p_deletion_to_deletion=convert_probability(state_switch_line[6]),
                p_emission_char=[convert_probability(p) for p in emission_line[1:-5]],
                p_insertion_char=[convert_probability(p) for p in insertion_line],
            )
        else:
            assert len(emission_line) == alphabet_length + 6, emission_line
            assert len(insertion_line) == alphabet_length, insertion_line
            assert len(state_switch_line) == 7, state_switch_line

            yield models.Step(
                p_emission_to_emission=convert_probability(state_switch_line[0]),
                p_emission_to_insertion=convert_probability(state_switch_line[1]),
                p_emission_to_deletion=convert_probability(state_switch_line[2]),
                p_insertion_to_emission=convert_probability(state_switch_line[3]),
                p_insertion_to_insertion=convert_probability(state_switch_line[4]),
                p_deletion_to_emission=convert_probability(state_switch_line[5]),
                p_deletion_to_deletion=convert_probability(state_switch_line[6]),
                p_emission_char=[convert_probability(p) for p in emission_line[1:-5]],
                p_insertion_char=[convert_probability(p) for p in insertion_line],
                alignment_column_index=None if emission_line[-1] == '-' else int(emission_line[-5]),
                consensus_residue_annotation=None if emission_line[-1] == '-' else emission_line[-4],
                reference_annotation=None if emission_line[-1] == '-' else emission_line[-3],
                mask_value=None if emission_line[-1] == '-' else emission_line[-2],
                annotation=None if emission_line[-1] == '-' else emission_line[-1],
            )


def convert_probability(p: str) -> float:
    if p == '*':
        return 0
    else:
        return math.e ** -float(p)
